var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _PostgresInstance_instances, _PostgresInstance_authMethod, _PostgresInstance_id, _PostgresInstance_options, _PostgresInstance_process, _PostgresInstance_isRootUser, _PostgresInstance_isInitialize, _PostgresInstance_hasPostgresStarted, _PostgresInstance_createDataDir, _PostgresInstance_createPasswordFile, _PostgresInstance_initDB, _PostgresInstance_checkForRootUser, _PostgresInstance_getUidAndGid;
import crypto from "node:crypto";
import { join } from "node:path";
import { spawn } from "node:child_process";
import pg from "pg";
const { Client } = pg;
import fs from "fs/promises";
import { existsSync } from "fs";
import { installPostgres, getPostgresBinaries } from "window-postgres-installer";
import execAsync from "./exec-async.js";
import { platform, tmpdir, userInfo } from "node:os";
class PostgresInstance {
    constructor(id, options) {
        _PostgresInstance_instances.add(this);
        _PostgresInstance_authMethod.set(this, "password");
        _PostgresInstance_id.set(this, void 0);
        _PostgresInstance_options.set(this, void 0);
        _PostgresInstance_process.set(this, void 0);
        _PostgresInstance_isRootUser.set(this, void 0);
        _PostgresInstance_isInitialize.set(this, false);
        _PostgresInstance_hasPostgresStarted.set(this, false);
        __classPrivateFieldSet(this, _PostgresInstance_id, id, "f");
        __classPrivateFieldSet(this, _PostgresInstance_options, options, "f");
        __classPrivateFieldSet(this, _PostgresInstance_isRootUser, userInfo().uid === 0, "f");
    }
    /**
     * Start the Postgres instance. It is automatically
     * shut down when the script exits.
     */
    start() {
        return __awaiter(this, void 0, void 0, function* () {
            yield installPostgres();
            let { postgres } = yield getPostgresBinaries();
            // Optionally retrieve the uid and gid
            const permissionIds = yield __classPrivateFieldGet(this, _PostgresInstance_instances, "m", _PostgresInstance_getUidAndGid).call(this)
                .catch(() => {
                throw new Error("Postgres cannot run as a root user. embedded-postgres could not find a postgres user to run as instead. Consider using the `createPostgresUser` option.");
            });
            // Greedily make the file executable, in case it is not
            yield fs.chmod(postgres, "755");
            yield new Promise((resolve, reject) => {
                var _a;
                // Spawn a postgres server
                __classPrivateFieldSet(this, _PostgresInstance_process, spawn(postgres, [
                    "-D",
                    __classPrivateFieldGet(this, _PostgresInstance_options, "f").databaseDir,
                    "-p",
                    __classPrivateFieldGet(this, _PostgresInstance_options, "f").port.toString(),
                    ...__classPrivateFieldGet(this, _PostgresInstance_options, "f").postgresFlags,
                ], Object.assign({}, permissionIds)), "f");
                (_a = __classPrivateFieldGet(this, _PostgresInstance_process, "f").stderr) === null || _a === void 0 ? void 0 : _a.on("data", (chunk) => {
                    const message = chunk.toString("utf-8");
                    __classPrivateFieldGet(this, _PostgresInstance_options, "f").onLog(message);
                    // Check for the right message to determine server start
                    if (message.includes("database system is ready to accept connections")) {
                        __classPrivateFieldSet(this, _PostgresInstance_hasPostgresStarted, true, "f");
                        resolve();
                    }
                });
                __classPrivateFieldGet(this, _PostgresInstance_process, "f").on("close", () => {
                    reject();
                });
            });
        });
    }
    /**
     * Stop the postgres instance.
     */
    stop() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!__classPrivateFieldGet(this, _PostgresInstance_process, "f")) {
                return;
            }
            if (!__classPrivateFieldGet(this, _PostgresInstance_hasPostgresStarted, "f")) {
                return;
            }
            // Kill the existing postgres process
            yield new Promise((resolve) => {
                var _a, _b, _c;
                // Register a handler for when the process finally exists
                (_a = __classPrivateFieldGet(this, _PostgresInstance_process, "f")) === null || _a === void 0 ? void 0 : _a.on("exit", resolve);
                if (platform() === "win32") {
                    if (!((_b = __classPrivateFieldGet(this, _PostgresInstance_process, "f")) === null || _b === void 0 ? void 0 : _b.pid)) {
                        throw new Error("Could not find process PID");
                    }
                    // Actually kill the process using the Windows taskkill command
                    spawn("taskkill", ["/pid", __classPrivateFieldGet(this, _PostgresInstance_process, "f").pid.toString(), "/f", "/t"]);
                }
                else {
                    (_c = __classPrivateFieldGet(this, _PostgresInstance_process, "f")) === null || _c === void 0 ? void 0 : _c.kill("SIGINT");
                }
            });
            // Additional work if database is not persistent
            if (__classPrivateFieldGet(this, _PostgresInstance_options, "f").persistent === false) {
                // Delete the data directory
                yield fs.rm(__classPrivateFieldGet(this, _PostgresInstance_options, "f").databaseDir, { recursive: true, force: true });
            }
            __classPrivateFieldSet(this, _PostgresInstance_process, undefined, "f");
            __classPrivateFieldSet(this, _PostgresInstance_hasPostgresStarted, false, "f");
            __classPrivateFieldSet(this, _PostgresInstance_isInitialize, __classPrivateFieldGet(this, _PostgresInstance_isInitialize, "f") && !__classPrivateFieldGet(this, _PostgresInstance_options, "f").persistent, "f");
        });
    }
    /**
     * Initialize the postgres instance.
     */
    initialize() {
        return __awaiter(this, void 0, void 0, function* () {
            if (__classPrivateFieldGet(this, _PostgresInstance_isInitialize, "f")) {
                return;
            }
            yield installPostgres();
            let { initdb } = yield getPostgresBinaries();
            yield __classPrivateFieldGet(this, _PostgresInstance_instances, "m", _PostgresInstance_checkForRootUser).call(this);
            let permissionIds = yield __classPrivateFieldGet(this, _PostgresInstance_instances, "m", _PostgresInstance_getUidAndGid).call(this)
                .catch(() => ({}));
            if (__classPrivateFieldGet(this, _PostgresInstance_options, "f").createPostgresUser && !("uid" in permissionIds) && !("gid" in permissionIds)) {
                try {
                    // Create the group and user
                    yield execAsync("groupadd postgres");
                    yield execAsync("useradd -g postgres postgres");
                    // Re-try the permission ids now the user exists
                    permissionIds = yield __classPrivateFieldGet(this, _PostgresInstance_instances, "m", _PostgresInstance_getUidAndGid).call(this);
                }
                catch (err) {
                    __classPrivateFieldGet(this, _PostgresInstance_options, "f").onError(err);
                    throw new Error("Failed to create and initialize a postgres user on this system.");
                }
            }
            if (__classPrivateFieldGet(this, _PostgresInstance_options, "f").createPostgresUser) {
                if (!('uid' in permissionIds)) {
                    throw new Error("Failed to retrieve the uid for the newly created user.");
                }
                yield __classPrivateFieldGet(this, _PostgresInstance_instances, "m", _PostgresInstance_createDataDir).call(this, permissionIds);
            }
            let passwordFile = yield __classPrivateFieldGet(this, _PostgresInstance_instances, "m", _PostgresInstance_createPasswordFile).call(this);
            yield __classPrivateFieldGet(this, _PostgresInstance_instances, "m", _PostgresInstance_initDB).call(this, initdb, passwordFile, permissionIds);
        });
    }
    /**
     * Create a database with a given name on the postgres instance.
     */
    createDatabase(databaseName) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!__classPrivateFieldGet(this, _PostgresInstance_process, "f")) {
                throw new Error('Your postgres instance must be running before you can create a database');
            }
            const client = this.client();
            yield client.connect();
            yield client.query(`CREATE DATABASE ${client.escapeIdentifier(databaseName)}`);
            yield client.end();
        });
    }
    /**
     * Make a query on the postgres instance.
     *
     * @param database The database that the postgres client should perform the query from
     * @param host The host that the postgres client should connect to
     */
    query(queryTextOrConfig_1) {
        return __awaiter(this, arguments, void 0, function* (queryTextOrConfig, values = [], database = "postgres", host = "localhost") {
            if (!__classPrivateFieldGet(this, _PostgresInstance_process, "f")) {
                throw new Error('Your postgres instance must be running before you can create a database');
            }
            const client = this.client(database, host);
            yield client.connect();
            const query = yield client.query(queryTextOrConfig, values);
            yield client.end();
            return query;
        });
    }
    /**
     * Determine if database with a given name on the postgres instance exist.
     */
    hasDatabase(databaseName) {
        return __awaiter(this, void 0, void 0, function* () {
            let dbQuery = yield this.query("SELECT * FROM  pg_catalog.pg_database WHERE datname=$1::text", [databaseName]);
            if (dbQuery.rowCount) {
                if (dbQuery.rowCount > 0) {
                    return true;
                }
            }
            return false;
        });
    }
    /**
     * Drop a database with a given name on the postgres instance.
     */
    dropDatabase(databaseName) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!__classPrivateFieldGet(this, _PostgresInstance_process, "f")) {
                throw new Error('Your postgres instance must be running before you can create a database');
            }
            const client = this.client();
            yield client.connect();
            yield client.query(`DROP DATABASE ${client.escapeIdentifier(databaseName)}`);
            yield client.end();
        });
    }
    /**
     * Create a node postgres client using the existing postgres instance configuration.
     *
     * @param database The database that the postgres client should connect to
     * @param host The host that should be pre-filled in the connection options
     * @returns Client
     */
    client(database = "postgres", host = "localhost") {
        // Create client
        const client = new Client({
            user: __classPrivateFieldGet(this, _PostgresInstance_options, "f").user,
            password: __classPrivateFieldGet(this, _PostgresInstance_options, "f").password,
            port: __classPrivateFieldGet(this, _PostgresInstance_options, "f").port,
            host,
            database,
        });
        // Log errors rather than throwing them so that embedded-postgres has
        // enough time to actually shutdown.
        client.on('error', __classPrivateFieldGet(this, _PostgresInstance_options, "f").onError);
        return client;
    }
    /**
     * Get the postgres instance id.
     */
    get id() {
        return __classPrivateFieldGet(this, _PostgresInstance_id, "f");
    }
}
_PostgresInstance_authMethod = new WeakMap(), _PostgresInstance_id = new WeakMap(), _PostgresInstance_options = new WeakMap(), _PostgresInstance_process = new WeakMap(), _PostgresInstance_isRootUser = new WeakMap(), _PostgresInstance_isInitialize = new WeakMap(), _PostgresInstance_hasPostgresStarted = new WeakMap(), _PostgresInstance_instances = new WeakSet(), _PostgresInstance_createDataDir = function _PostgresInstance_createDataDir(permissionIds) {
    return __awaiter(this, void 0, void 0, function* () {
        yield fs.mkdir(__classPrivateFieldGet(this, _PostgresInstance_options, "f").databaseDir, { recursive: true });
        yield fs.chown(__classPrivateFieldGet(this, _PostgresInstance_options, "f").databaseDir, permissionIds.uid, permissionIds.gid);
    });
}, _PostgresInstance_createPasswordFile = function _PostgresInstance_createPasswordFile() {
    return __awaiter(this, void 0, void 0, function* () {
        const randomId = crypto.randomBytes(6).readUIntLE(0, 6).toString(36);
        const passwordFile = join(tmpdir(), `pg-password-${randomId}`);
        yield fs.writeFile(passwordFile, __classPrivateFieldGet(this, _PostgresInstance_options, "f").password + "\n");
        return passwordFile;
    });
}, _PostgresInstance_initDB = function _PostgresInstance_initDB(initdbExe, passwordFile, permissionIds) {
    return __awaiter(this, void 0, void 0, function* () {
        if (existsSync(__classPrivateFieldGet(this, _PostgresInstance_options, "f").databaseDir)) {
            if (__classPrivateFieldGet(this, _PostgresInstance_options, "f").overwriteDatabaseDir) {
                yield fs.rm(__classPrivateFieldGet(this, _PostgresInstance_options, "f").databaseDir, { recursive: true, force: true });
            }
            else {
                throw new Error(`Directory ${__classPrivateFieldGet(this, _PostgresInstance_options, "f").databaseDir} exists`);
            }
        }
        // Greedily make the file executable, in case it is not
        yield fs.chmod(initdbExe, "755");
        yield new Promise((resolve, reject) => {
            const process = spawn(initdbExe, [
                `--pgdata=${__classPrivateFieldGet(this, _PostgresInstance_options, "f").databaseDir}`,
                `--auth=${__classPrivateFieldGet(this, _PostgresInstance_authMethod, "f")}`,
                `--username=${__classPrivateFieldGet(this, _PostgresInstance_options, "f").user}`,
                `--pwfile=${passwordFile}`,
                ...__classPrivateFieldGet(this, _PostgresInstance_options, "f").initdbFlags,
            ], Object.assign({ stdio: "inherit" }, permissionIds));
            process.on("exit", (code) => {
                if (code === 0) {
                    resolve();
                }
                else {
                    reject(`Postgres init script exited with code ${code}. Please check the logs for extra info. The data directory might already exist.`);
                }
            });
        });
        yield fs.unlink(passwordFile);
        __classPrivateFieldSet(this, _PostgresInstance_isInitialize, true, "f");
    });
}, _PostgresInstance_checkForRootUser = function _PostgresInstance_checkForRootUser() {
    return __awaiter(this, void 0, void 0, function* () {
        if (!__classPrivateFieldGet(this, _PostgresInstance_isRootUser, "f")) {
            return;
        }
        // Attempt to retrieve the uid and gid for the postgres user. This check
        // will throw and error when the postgres user doesn't exist
        try {
            yield __classPrivateFieldGet(this, _PostgresInstance_instances, "m", _PostgresInstance_getUidAndGid).call(this);
        }
        catch (err) {
            // No user exists, but check that a postgres user should be created
            if (!__classPrivateFieldGet(this, _PostgresInstance_options, "f").createPostgresUser) {
                throw new Error("You are running this script as root. Postgres does not support running as root. If you wish to continue, configure postgres instance to create a Postgres user by setting the `createPostgresUser` option to true.");
            }
        }
    });
}, _PostgresInstance_getUidAndGid = function _PostgresInstance_getUidAndGid() {
    return __awaiter(this, arguments, void 0, function* (name = "postgres") {
        if (!__classPrivateFieldGet(this, _PostgresInstance_isRootUser, "f")) {
            return {};
        }
        const [uid, gid] = yield Promise.all([
            execAsync(`id -u ${name}`).then(Number.parseInt),
            execAsync(`id -g ${name}`).then(Number.parseInt),
        ]);
        return { uid, gid };
    });
};
export default PostgresInstance;
